@@ -1,25 +1,29 @@
 <?php
 /**
- * Horde_Cli API for basic command-line functionality/checks.
- *
- * Copyright 2003-2016 Horde LLC (http://www.horde.org/)
+ * Copyright 2003-2017 Horde LLC (http://www.horde.org/)
  *
  * See the enclosed file COPYING for license information (LGPL). If you
  * did not receive this file, see http://www.horde.org/licenses/lgpl21.
  *
- * @author  Chuck Hagenbuch <chuck@horde.org>
- * @author  Jan Schneider <jan@horde.org>
- * @package Cli
+ * @author   Chuck Hagenbuch <chuck@horde.org>
+ * @author   Jan Schneider <jan@horde.org>
+ * @category Horde
+ * @license  http://www.horde.org/licenses/lgpl21 LGPL
+ * @package  Cli
+ */
+
+/**
+ * Horde_Cli API for basic command-line functionality/checks.
+ *
+ * @author    Chuck Hagenbuch <chuck@horde.org>
+ * @author    Jan Schneider <jan@horde.org>
+ * @category  Horde
+ * @copyright 2003-2017 Horde LLC
+ * @license   http://www.horde.org/licenses/lgpl21 LGPL
+ * @package   Cli
  */
 class Horde_Cli
 {
-    /**
-     * Are we running on a console?
-     *
-     * @var boolean
-     */
-    protected $_console;
-
     /**
      * The newline string to use.
      *
@@ -27,6 +31,13 @@ class Horde_Cli
      */
     protected $_newline;
 
+    /**
+     * The formatted space string to use.
+     *
+     * @var string
+     */
+    protected $_space;
+
     /**
      * The string to use for clearing the screen.
      *
@@ -42,81 +53,18 @@ class Horde_Cli
     protected $_indent;
 
     /**
-     * The string to mark the beginning of bold text.
+     * The color formatter.
      *
-     * @var string
+     * @var Horde_Cli_Color
      */
-    protected $_bold_start = '';
+    protected $_color;
 
     /**
-     * The string to mark the end of bold text.
+     * The terminal width.
      *
-     * @var string
+     * @var integer
      */
-    protected $_bold_end = '';
-
-    /**
-     * The strings to mark the beginning of coloured text.
-     *
-     * @var string
-     */
-    protected $_red_start    = '';
-    protected $_green_start  = '';
-    protected $_yellow_start = '';
-    protected $_blue_start   = '';
-
-    /**
-     * The strings to mark the end of coloured text.
-     *
-     * @var string
-     */
-    protected $_red_end      = '';
-    protected $_green_end    = '';
-    protected $_yellow_end   = '';
-    protected $_blue_end     = '';
-
-    /**
-     * Terminal foreground color codes. Not used yet.
-     *
-     * @var array
-     */
-    protected $_terminalForegrounds = array(
-        'normal'        => "\x1B[0m",
-        'black'         => "\x1B[0m",
-        'bold'          => "\x1b[1m",
-        'red'           => "\x1B[31m",
-        'green'         => "\x1B[32m",
-        'brown'         => "\x1B[33m",
-        'blue'          => "\x1B[34m",
-        'magenta'       => "\x1B[35m",
-        'cyan'          => "\x1B[36m",
-        'lightgray'     => "\x1B[37m",
-        'white'         => "\x1B[1m\x1B[37m",
-        'darkgray'      => "\x1B[1m\x1B[0m",
-        'lightred'      => "\x1B[1m\x1B[31m",
-        'lightgreen'    => "\x1B[1m\x1B[32m",
-        'yellow'        => "\x1B[1m\x1B[33m",
-        'lightblue'     => "\x1B[1m\x1B[34m",
-        'lightmagenta'  => "\x1B[1m\x1B[35m",
-        'lightcyan'     => "\x1B[1m\x1B[36m",
-    );
-
-    /**
-     * Terminal background color codes. Not used yet.
-     *
-     * @var array
-     */
-    protected $_terminalBackgrounds = array(
-        'normal'    => "\x1B[0m",
-        'black'     => "\x1B[0m",
-        'red'       => "\x1B[41m",
-        'green'     => "\x1B[42m",
-        'brown'     => "\x1B[43m",
-        'blue'      => "\x1B[44m",
-        'magenta'   => "\x1B[45m",
-        'cyan'      => "\x1B[46m",
-        'lightgray' => "\x1B[47m",
-    );
+    protected $_width;
 
     /**
      * Detect the current environment (web server or console) and sets
@@ -127,48 +75,43 @@ class Horde_Cli
      */
     public function __construct()
     {
-        $this->_console = $this->runningFromCLI();
+        $this->_color = new Horde_Cli_Color();
+        $console = $this->runningFromCLI();
 
-        if ($this->_console) {
-            $this->_newline = "\n";
-            $this->_indent  = '    ';
-
-            $term = getenv('TERM');
-            if ($term) {
-                if (preg_match('/^(xterm|vt220|linux)/', $term)) {
-                    $this->_clearscreen  = "\x1b[2J\x1b[H";
-                    $this->_bold_start   = "\x1b[1m";
-                    $this->_red_start    = "\x1b[01;31m";
-                    $this->_green_start  = "\x1b[01;32m";
-                    $this->_yellow_start = "\x1b[01;33m";
-                    $this->_blue_start   = "\x1b[01;34m";
-                    $this->_bold_end = $this->_red_end = $this->_green_end = $this->_yellow_end = $this->_blue_end = "\x1b[0m";
-                } elseif (preg_match('/^vt100/', $term)) {
-                    $this->_clearscreen  = "\x1b[2J\x1b[H";
-                    $this->_bold_start = "\x1b[1m";
-                    $this->_bold_end   = "\x1b[0m";
-                }
+        if ($console) {
+            $this->_newline = PHP_EOL;
+            $this->_space = ' ';
+            if (getenv('TERM')) {
+                $this->_clearscreen  = "\x1b[2J\x1b[H";
             }
+            $this->_setWidth();
         } else {
             $this->_newline = '<br />';
-            $this->_indent  = str_repeat('&nbsp;', 4);
-
-            $this->_bold_start   = '<strong>';
-            $this->_bold_end     = '</strong>';
-            $this->_red_start    = '<span style="color:red">';
-            $this->_green_start  = '<span style="color:green">';
-            $this->_yellow_start = '<span style="color:yellow">';
-            $this->_blue_start   = '<span style="color:blue">';
-            $this->_red_end = $this->_green_end = $this->_yellow_end = $this->_blue_end = '</span>';
+            $this->_space = '&nbsp;';
         }
+        $this->_indent = str_repeat($this->_space, 4);
 
         // We really want to call this at the end of the script, not in the
         // destructor.
-        if ($this->_console) {
+        if ($console) {
             register_shutdown_function(array($this, 'shutdown'));
         }
     }
 
+    /**
+     * Retuns the detected terminal screen width.
+     *
+     * Defaults to 80 if the width cannot be detected automatically.
+     *
+     * @since Horde_Cli 2.2.0
+     *
+     * @return integer  The terminal screen width or null if not a terminal.
+     */
+    public function getWidth()
+    {
+        return $this->_width;
+    }
+
     /**
      * Prints $text on a single line.
      *
@@ -208,61 +151,116 @@ class Horde_Cli
     /**
      * Returns a bold version of $text.
      *
+     * @deprecated Use Horde_Cli_Color instead.
+     *
      * @param string $text  The text to bold.
      *
      * @return string  The bolded text.
      */
     public function bold($text)
     {
-        return $this->_bold_start . $text . $this->_bold_end;
+        return $this->_color->bold($text);
+    }
+
+    /**
+     * Returns a colored version of $text.
+     *
+     * @since Horde_Cli 2.1.0
+     *
+     * @param string $color  The color to use. @see $_foregroundColors
+     * @param string $text   The text to print in this color.
+     *
+     * @return string  The colored text.
+     */
+    public function color($color, $text)
+    {
+        return $this->_color->color($color, $text);
     }
 
     /**
      * Returns a red version of $text.
      *
+     * @deprecated Use Horde_Cli_Color or color() instead.
+     *
      * @param string $text  The text to print in red.
      *
      * @return string  The red text.
      */
     public function red($text)
     {
-        return $this->_red_start . $text . $this->_red_end;
+        return $this->_color->red($text);
     }
 
     /**
      * Returns a green version of $text.
      *
+     * @deprecated Use Horde_Cli_Color or color() instead.
+     *
      * @param string $text  The text to print in green.
      *
      * @return string  The green text.
      */
     public function green($text)
     {
-        return $this->_green_start . $text . $this->_green_end;
+        return $this->_color->green($text);
     }
 
     /**
      * Returns a blue version of $text.
      *
+     * @deprecated Use Horde_Cli_Color or color() instead.
+     *
      * @param string $text  The text to print in blue.
      *
      * @return string  The blue text.
      */
     public function blue($text)
     {
-        return $this->_blue_start . $text . $this->_blue_end;
+        return $this->_color->blue($text);
     }
 
     /**
      * Returns a yellow version of $text.
      *
+     * @deprecated Use Horde_Cli_Color or color() instead.
+     *
      * @param string $text  The text to print in yellow.
      *
      * @return string  The yellow text.
      */
     public function yellow($text)
     {
-        return $this->_yellow_start . $text . $this->_yellow_end;
+        return $this->_color->yellow($text);
+    }
+
+    /**
+     * Creates a header from a string by drawing character lines above or below
+     * the header content.
+     *
+     * @since Horde_Cli 2.1.0
+     *
+     * @param string $message  A message to turn into a header.
+     * @param string $below    Character to use for drawing the line below the
+     *                         message.
+     * @param string $above    Character to use for drawing the line above the
+     *                         message.
+     */
+    public function header($message, $below = '-', $above = null)
+    {
+        $length = 0;
+        foreach (explode($this->_newline, $this->_color->remove($message)) as $line) {
+            $length = max($length, Horde_String::length($line));
+        }
+        if ($width = $this->getWidth()) {
+            $length = min($length, $width);
+        }
+        if (strlen($above)) {
+            $this->writeln(str_repeat($above, $length));
+        }
+        $this->writeln(wordwrap($message, $length, $this->_newline));
+        if (strlen($below)) {
+            $this->writeln(str_repeat($below, $length));
+        }
     }
 
     /**
@@ -274,31 +272,73 @@ class Horde_Cli
      */
     public function message($message, $type = 'cli.message')
     {
-        $message = wordwrap(str_replace("\n", "\n           ", $message),
-                            68, "\n           ", true);
+        if ($width = $this->getWidth()) {
+            $indent = $this->_newline . str_repeat($this->_space, 11);
+            $message = wordwrap(
+                str_replace($this->_newline, $indent, $message),
+                $width - 12,
+                $indent,
+                true
+            );
+        }
 
         switch ($type) {
         case 'cli.error':
-            $type_message = $this->red('[ ERROR! ] ');
+            $this->writeln(
+                $this->_color->lightgray(
+                    $this->block(
+                        '[' . $this->_space . 'ERROR!' . $this->_space . '] '
+                            . $message,
+                        'red'
+                    )
+                )
+            );
             break;
 
         case 'cli.warning':
-            $type_message = $this->yellow('[  WARN  ] ');
+            $this->writeln(
+                $this->_color->black(
+                    $this->block(
+                        '[' . $this->_space . $this->_space . 'WARN'
+                            . $this->_space . $this->_space . '] '
+                            . $message,
+                        'brown'
+                    )
+                )
+            );
             break;
 
         case 'cli.success':
-            $type_message = $this->green('[   OK   ] ');
+            $this->writeln(
+                $this->_color->black(
+                    $this->block(
+                        '[' . $this->_space . $this->_space . $this->_space
+                            . 'OK' . $this->_space . $this->_space
+                            . $this->_space . '] '
+                            . $message,
+                        'green'
+                    )
+                )
+            );
             break;
 
         case 'cli.message':
-            $type_message = $this->blue('[  INFO  ] ');
+            $this->writeln(
+                $this->_color->lightgray(
+                    $this->block(
+                        '[' . $this->_space . $this->_space . 'INFO'
+                            . $this->_space . $this->_space . '] '
+                            . $message,
+                        'blue'
+                    )
+                )
+            );
             break;
 
         default:
-            $type_message = '';
+            $this->writeln($message);
+            break;
         }
-
-        $this->writeln($type_message . $message);
     }
 
     /**
@@ -309,7 +349,8 @@ class Horde_Cli
      */
     public function fatal($error)
     {
-        if ($error instanceof Exception) {
+        if ($error instanceof Throwable ||
+            $error instanceof Exception) {
             $trace = $error;
         } else {
             $trace = debug_backtrace();
@@ -342,23 +383,83 @@ class Horde_Cli
             }
             $error = $error->getMessage();
         }
-        $this->writeln();
-        $this->writeln($this->red('===================='));
-        $this->writeln();
-        $this->writeln($this->red(Horde_Cli_Translation::t("Fatal Error:")));
-        $this->writeln($this->red($error));
+
+        $lines = array('');
+        $lines = $this->_addLines(
+            $lines, Horde_Cli_Translation::t("Fatal Error:")
+        );
+        $lines = $this->_addLines($lines, $error);
         if ($details) {
-            $this->writeln($this->red(print_r($details, true)));
+            $this->_addLines($lines, print_r($details, true));
         }
         if ($location) {
-            $this->writeln($this->red($location));
+            $lines = $this->_addLines($lines, $location);
         }
-        $this->writeln();
-        $this->writeln((string)$backtrace);
-        $this->writeln($this->red('===================='));
+        $lines = $this->_addLines($lines, '');
+        $lines = $this->_addLines($lines, (string)$backtrace);
+        $lines = $this->_addLines($lines, '');
+        $this->writeln(
+            $this->_color->lightgray(
+                $this->block(implode($this->_newline, $lines), 'red', '  ')
+            )
+        );
         exit(1);
     }
 
+    /**
+     * Adds content to a line buffer, wrapping long lines if necessary.
+     *
+     * @param array $buffer    The line buffer.
+     * @param string $content  The lines to add.
+     *
+     * @return string  The updated line buffer.
+     */
+    protected function _addLines($buffer, $content)
+    {
+        $width = max(0, $this->getWidth() - 7);
+        if ($width) {
+            foreach (explode($this->_newline, $content) as $line) {
+                $buffer[] = wordwrap($line, $width, "\n   ", true);
+            }
+        } else {
+            $buffer[] = $content;
+        }
+        return $buffer;
+    }
+
+    /**
+     * Formats text in a visual block with optional margin.
+     *
+     * @since Horde_Cli 2.2.0
+     *
+     * @param string $text    The block text.
+     * @param string $color   The background color.
+     * @param string $margin  The block margin string.
+     * @param integer $width  The block width.
+     *
+     * @return string  The formatted block.
+     */
+    public function block($text, $color, $margin = '', $width = null)
+    {
+        $text = explode($this->_newline, $text);
+        if (!$width) {
+            $width = 0;
+            foreach ($text as $line) {
+                $width = max($width, strlen($line));
+            }
+            if ($maxWidth = $this->getWidth()) {
+                $width = min($width, $maxWidth - 2 * strlen($margin));
+            }
+        }
+        foreach ($text as &$line) {
+            $line = $this->_color->background(
+                $color,
+                $margin . sprintf('%-' . $width . 's', $line) . $margin
+            );
+        }
+        return implode($this->_newline, $text);
+    }
+
     /**
      * Prompts for a user response.
      *
@@ -373,46 +474,60 @@ class Horde_Cli
      */
     public function prompt($prompt, $choices = null, $default = null)
     {
+        $width = $this->getWidth();
+
+        if (!is_array($choices) || empty($choices)) {
+            if ($default !== null) {
+                $prompt .= ' [' . $default . ']';
+            }
+            if ($width) {
+                $prompt = wordwrap($prompt, $width);
+            }
+            $this->writeln($prompt . ' ', true);
+            @ob_flush();
+            $response = trim(fgets(STDIN));
+            if ($response === '' && $default !== null) {
+                $response = $default;
+            }
+            return $response;
+        }
+
+        if ($width) {
+            $prompt = wordwrap($prompt, $width);
+        }
+
         // Main event loop to capture top level command.
         while (true) {
             // Print out the prompt message.
-            if (is_array($choices) && !empty($choices)) {
-                $this->writeln(wordwrap($prompt) . ' ', !is_array($choices));
-                foreach ($choices as $key => $choice) {
-                    $this->writeln($this->indent('(' . $this->bold($key) . ') ' . $choice));
-                }
-                $question = Horde_Cli_Translation::t("Type your choice");
-                if ($default !== null) {
-                    $question .= ' [' . $default . ']';
-                }
-                $this->writeln($question . ': ', true);
-                @ob_flush();
+            $this->writeln($prompt . ' ', !is_array($choices));
+            foreach ($choices as $key => $choice) {
+                $this->writeln(
+                    $this->indent('(' . $this->_color->bold($key) . ') ' . $choice)
+                );
+            }
+            $question = Horde_Cli_Translation::t("Type your choice");
+            if ($default !== null) {
+                $question .= ' [' . $default . ']';
+            }
+            $this->writeln($question . ': ', true);
+            @ob_flush();
 
-                // Get the user choice.
-                $response = trim(fgets(STDIN));
-                if ($response === '' && $default !== null) {
-                    $response = $default;
-                }
-                if (isset($choices[$response])) {
-                    return $response;
-                } else {
-                    $this->writeln($this->red(sprintf(Horde_Cli_Translation::t("\"%s\" is not a valid choice."), $response)));
-                }
-            } else {
-                if ($default !== null) {
-                    $prompt .= ' [' . $default . ']';
-                }
-                $this->writeln(wordwrap($prompt) . ' ', true);
-                @ob_flush();
-                $response = trim(fgets(STDIN));
-                if ($response === '' && $default !== null) {
-                    $response = $default;
-                }
+            // Get the user choice.
+            $response = trim(fgets(STDIN));
+            if ($response === '' && $default !== null) {
+                $response = $default;
+            }
+            if (isset($choices[$response])) {
                 return $response;
+            } else {
+                $this->writeln($this->_color->red(sprintf(
+                    Horde_Cli_Translation::t(
+                        "\"%s\" is not a valid choice."
+                    ),
+                    $response
+                )));
             }
         }
-
-        return true;
     }
 
     /**
@@ -432,7 +547,11 @@ class Horde_Cli
 
         if (preg_match('/^win/i', PHP_OS)) {
             $vbscript = sys_get_temp_dir() . 'prompt_password.vbs';
-            file_put_contents($vbscript, 'wscript.echo(InputBox("' . addslashes($prompt) . '", "", "password here"))');
+            file_put_contents(
+                $vbscript,
+                'wscript.echo(InputBox("' . addslashes($prompt)
+                . '", "", "password here"))'
+            );
             $command = "cscript //nologo " . escapeshellarg($vbscript);
             $password = rtrim(shell_exec($command));
             unlink($vbscript);
@@ -442,17 +561,18 @@ class Horde_Cli
                 /* Cannot spawn shell, fall back to standard prompt. */
                 return $this->prompt($prompt);
             }
-            $command = '/usr/bin/env bash -c "read -s -p ' . escapeshellarg($prompt) . ' mypassword && echo \$mypassword"';
+            $command = '/usr/bin/env bash -c "read -s -p '
+                . escapeshellarg($prompt) . ' mypassword && echo \$mypassword"';
             $password = rtrim(shell_exec($command));
-            echo "\n";
+            echo $this->_newline;
         }
 
         return $password;
     }
 
     /**
-     * Reads everything that is sent through standard input and returns it
-     * as a single string.
+     * Reads everything that is sent through standard input and returns it as a
+     * single string.
      *
      * @return string  The contents of the standard input.
      */
@@ -536,4 +656,23 @@ class Horde_Cli
         }
     }
 
+    /**
+     * Detects the terminal screen width.
+     */
+    protected function _setWidth()
+    {
+        $this->_width = getenv('COLUMNS');
+        if (!$this->_width) {
+            $this->_width = @exec('tput cols 2> /dev/null');
+        }
+        if (!$this->_width) {
+            $size = explode(' ', @exec('stty size 2> /dev/null'));
+            if (count($size) == 2) {
+                $this->_width = $size[1];
+            }
+        }
+        if (!$this->_width) {
+            $this->_width = 80;
+        }
+    }
 }
